import random
import statistics
def mediane(L):
    """ Calcul la médiane de L.

    m := mediane(L) est un entier tel que:
        le nombre d'entier plus petit que m dans L
        est égale (+/- 1) au nombre d'entier plus grand.

    Complexité:
    -----------
    Deux boucle imbriqué sur L, en posant n := O(len(L)), on a une complexité en O(n^2).

    Example:
    --------

    >>> mediane([0, 2, 3, 4, 5])
    3
    >>> mediane([5, 4, 3, 0, 2])
    3
    >>> m = mediane([11, 5, 3, 2, 23, 50]) # n'importe quel nombre entre 5 et 11
    >>> assert(m >= 5 and m <= 11)
    >>> random.seed(4242)
    >>> L = random.choices(list(range(100)), k=123)
    >>> assert(mediane(L) == statistics.median(L))
    >>> L = random.choices(list(range(100)), k=124)
    >>> assert(mediane(L) == statistics.median(L))

    """
    d = {}  # On stock l'information du nombre d'élément plus petit.
    occ = {}
    n = len(L)/2
    for e in L:
        d[e] = 0
        occ[e] = 0
        for f in L:
            if f < e:
                d[e] += 1
            if e == f:
                occ[e] += 1
    for e in L:
        if d[e] <= n and n <= d[e] + occ[e]:
            return e

def mediane_sort(L):
    """
    Complexité:
    -----------
    sorted(L) est de complexité O(nlog(n)), donc complexité en O(nlog(n))

    Example:
    --------

    >>> mediane_sort([0, 2, 3, 4, 5])
    3
    >>> mediane_sort([5, 4, 3, 0, 2])
    3
    >>> m = mediane_sort([11, 5, 3, 2, 23, 50]) # n'importe quel nombre entre 5 et 11
    >>> assert(m >= 5 and m <= 11)

    """
    return sorted(L)[len(L)//2]

def selection(L, k):
    """
    Retourne l'entier en position "k" dans la liste triée.

    Complexité:
    -----------

    Même complexité que le trie, donc en O(nlog(n)).
    """
    return sorted(L)[k]

def selection_imperative(L, k):
    """
    Retourne l'entier en position "k" dans la liste triée.

    La version impérative généralise le calcul du min.

    Complexité:
    -----------
    On fait deux boucles imbriqués une de taille n et une de k.
    On a une complexité en O(nk)
    """
    min_indices = []
    for i in range(k):
        min_indices.append(-1)
    # On initialise la liste d'indice à -1
    n = len(L)
    for i in range(n):
        element = L[i]
        insertion = False # est-ce que l'indice doit être inséré?
        indice = -1 # à quel position
        if min_indices[0] == -1:
            min_indices[0] = i
        for j in range(k-1):
            # Si le tableau est pas initialisé au niveau j, on l'initialise et on quitte.
            if min_indices[j] == -1:
                min_indices[j] = i
                break
            # le tableau peut ne pas être initialisé en j+1
            # if faut insérer soit en position j soit en position j+1.
            if min_indices[j+1] == -1:
                if min_indices[j] < element:
                    min_indices[j+1] = i
                else:
                    min_indices[j+1] = min_indices[j]
                    min_indices[j] = i
                break
            # le cas général ici.
            if L[min_indices[j]] < element and L[min_indices[j+1]] > element and not insertion:
                # on a trouvé un indice a insérer et une position.
                insertion = True
                indice = j
            if insertion:
                # on doit décalé à droite le reste des indices
                min_indices[j+1] = min_indice[j]
        if insertion:
            min_indice[i] = element



def selection_rec(L, k):
    """
        Retourne l'entier en position "k" dans la liste triée.

        Version récursive
        -----------------
        Inspiré par la page Wikipédia:
            - https://en.wikipedia.org/wiki/Selection_algorithm
            - https://fr.wikipedia.org/wiki/Algorithme_de_s%C3%A9lection (Français)
    """
        if len(L) <= 5:
            return selection(L, k)
        n = len(L)//5
        K = [] # la liste des médianes des tableau de taille 5
        for i in range(n):
            K.append(selection_rec(L[5*i:5*(i+1)], k=2))
        m = selection_rec(K, k=len(K)//2) # m est la médiane des médiane.
        au_dessus = []
        en_dessous = []
        for e in L:
            if e < m:
                en_dessous.append(e)
            else:
                au_dessus.append(e)
        if len(en_dessous) <= k:
            return select_rec(au_dessus, k-len(en_dessous))
        else:
            return select_rec(en_dessous, k)

def mediane_rec(L):
    return selection_rec(L, len(L)//2)
